/* TroiaPreview
 * version 0.2 date 2004-09-01
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <windows.h>

#include "Tools.h"
#include "Vector.h"
#include "Images.h"
#include "Log.h"
#include "Textures.h"
#include "Console.h"
#include "Globals.h"
#include "Pixel.h"
#include "Frame.h"
#include "Commands.h"

//private constants
#define ScriptTimerTime 20
#define ScriptBufferSizeMax 1048576 //maximum size of script-buffer
#define MaxScriptCommandsAtSameTime 20 //maximum number of short command that are executed at the same time

//private global variables
char ScriptMode = 'R'; //S=stopped, W=waiting, R=running
double ScriptWaitTime = 0.0; //time to wait for in ScriptMode is W
char * ScriptBuffer = NULL; //buffer with script-lines to process

//load a script-file
static long LoadScript( char * FileName )
{
	//local variables
	FILE * File;
	char Line[1024];
	char * pChr;
	long I, L, Cnt;

	//script-buffer too large
	if( strlen( ScriptBuffer ) >= ScriptBufferSizeMax )
		return -1;
	
	//empty filename ===> fopen will raise assertion, so exit
	if( FileName == NULL || FileName[0] == 0 )
		return -1;

	//open file
	File = fopen( FileName, "rt" );
	if( File == NULL )
		return -1;

	//script-buffer not empty
	if( ScriptBuffer[0] != 0 )
	{
		//script-buffer does not end with newline
		L = strlen( ScriptBuffer );
		if( ScriptBuffer[L - 1] != '\n' )
		{
			//add a newline
			ScriptBuffer = (char*)realloc( ScriptBuffer, L + 2 );
			ScriptBuffer[L] = '\n';
			ScriptBuffer[L + 1] = 0;
		}
	}
	
	//read entire file line by line
	Cnt = 0;
	while( !feof( File ) )
	{
		//read a line
		Line[0] = 0;
		fgets( Line, sizeof( Line ), File );
		//replace newline with 0
		pChr = strchr( Line, '\n' );
		if( pChr != NULL )
			*pChr = 0;
		//remove illegal characters from line
		L = strlen( Line );
		for( I = 0; Line[I] != 0; I++ )
			if( (unsigned char)Line[I] < ' ' )
				memmove( Line + I, Line + I + 1, L - I );
		//when line not empty
		if( Line[0] != 0 )
		{
			//add line and newline to script-buffer
			L = strlen( ScriptBuffer );
			ScriptBuffer = (char*)realloc( ScriptBuffer, L + strlen( Line ) + 2 );
			strcpy( ScriptBuffer + L, Line );
			strcat( ScriptBuffer + L, "\n" );
			//count lines
			Cnt++;
		}
	}

	//close file
	fclose( File );

	//return number of lines
	return Cnt;
}

//process a command
//return true for "long" commands, false for "short" commands
int ProcessCommand( char * Command )
{
	//local variables
	char * pChr, * pVal;
	long I, J, K;
	char Text[256];
	bool FuRes;
	unsigned char Color, Color2, Dir;
	tPixel3D Pixel;
	unsigned short ForeColor, BackColor;

	//help
	if( strnicmp( Command, "help", 4 ) == 0 )
	{
		//display help
		ConsoleInsertLine( "  Exit, Quit: end program" );
		ConsoleInsertLine( "  Start: go to startpoint" );
		ConsoleInsertLine( "  Pos [x=<x>] [y=<y>] [z=<z>]: set or view position" );
		ConsoleInsertLine( "  Front [x=<x>] [y=<y>] [z=<z>]: set or view front-direction" );
		ConsoleInsertLine( "  CamptShow: show cam-points" );
		ConsoleInsertLine( "  CamptClr: clear cam-points" );
		ConsoleInsertLine( "  CamptSet <n> [px=<>] [py=<>] [pz=<>] [fx=<>] [fy=<>] [fz=<>]: set cam-point n" );
		ConsoleInsertLine( "  CamptIns <n> [px=<>] [py=<>] [pz=<>] [fx=<>] [fy=<>] [fz=<>]: insert cam-point n" );
		ConsoleInsertLine( "    p?: position of cam, f?: front direction of cam" );
		ConsoleInsertLine( "  CamptDel <n>: delete cam-point n" );
		ConsoleInsertLine( "  CamptSpeed [<speed>]: waypoint-move-speed in cam-points per second" );
		ConsoleInsertLine( "  CamptRefTime [Time|<seconds>]: reference time for cam-movement" );
		ConsoleInsertLine( "  CamMove [on|off]: enable / disable movement of camera" );
		ConsoleInsertLine( "  Console [on|off]: show / hide the console" );
		ConsoleInsertLine( "  BackGr [on|off]: enable / disable background" );
		ConsoleInsertLine( "  AntiAlias [on|off]: enable / disable antialiasing" );
		ConsoleInsertLine( "  BackLoadBmp [<filename>]: load background from bmp-file" );
		ConsoleInsertLine( "  BackLoadJpg [<filename>]: load background from jpg-file" );
		ConsoleInsertLine( "  BackUnload: unload background" );
		ConsoleInsertLine( "  PixTexLoadBmp [<filename>]: load pixel texture from bmp-file" );
		ConsoleInsertLine( "  PixTexLoadJpg [<filename>]: load pixel texture from jpg-file" );
		ConsoleInsertLine( "  PixTexUnload: unload pixel texture" );
		ConsoleInsertLine( "  Time [<seconds>]: get / set time" );
		ConsoleInsertLine( "  ScriptLoad <filename>: load a script-file with commands" );
		ConsoleInsertLine( "  ScriptUnload: stop script execution and forget loaded scripts" );
		ConsoleInsertLine( "  ScriptStart: go on with script execution" );
		ConsoleInsertLine( "  ScriptStop: halt script execution" );
		ConsoleInsertLine( "  ScriptWaitAbs <seconds>: pause script execution until specified time reached" );
		ConsoleInsertLine( "  ScriptWaitRel <seconds>: pause script execution for specified number of seconds" );
		ConsoleInsertLine( "  ColorSet [<color>]: set entire room to a color" );
		ConsoleInsertLine( "  ColorRandom [<min color>] [<max color>]: set entire room to random colors" );
		ConsoleInsertLine( "  TextWrite <pixel> <direction> <foreground> <background> <text>: write text to the room" );
		ConsoleInsertLine( "    pixel: e.g. W1/0/0, direction: e.g. right, color 256 is transparent" );
		//long command
		return true;
	}

	//end program
	if( strnicmp( Command, "exit", 4 ) == 0 || strnicmp( Command, "quit", 4 ) == 0 )
	{
		PostQuitMessage( 0 );
		//long command
		return true;
	}

	//start-position
	if( strnicmp( Command, "start", 5 ) == 0 )
	{
		SetVector( &Pos, CenterX, CenterY, CenterZ );
		SetVector( &Front, 0.0, 1.0, 0.0 );
		//short command
		return false;
	}

	//go to position
	if( strnicmp( Command, "pos", 3 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 3;
		for( ; *pChr == ' '; pChr++ );
		//new x-value
		pVal = stristr( pChr, "x=" );
		if( pVal != NULL )
			Pos.X = DoubleMinMax( atof( pVal + 2 ), PosMinX, PosMaxX );
		//new y-value
		pVal = stristr( pChr, "y=" );
		if( pVal != NULL )
			Pos.Y = DoubleMinMax( atof( pVal + 2 ), PosMinY, PosMaxY );
		//new z-value
		pVal = stristr( pChr, "z=" );
		if( pVal != NULL )
			Pos.Z = DoubleMinMax( atof( pVal + 2 ), PosMinZ, PosMaxZ );
		//display position
		sprintf( Text, "  position is (%f, %f, %f)", Pos.X, Pos.Y, Pos.Z );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//set front-direction
	if( strnicmp( Command, "front", 5 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 5;
		for( ; *pChr == ' '; pChr++ );
		//new x-value
		pVal = stristr( pChr, "x=" );
		if( pVal != NULL )
			Front.X = DoubleMinMax( atof( pVal + 2 ), -1.0, 1.0 );
		//new y-value
		pVal = stristr( pChr, "y=" );
		if( pVal != NULL )
			Front.Y = DoubleMinMax( atof( pVal + 2 ), -1.0, 1.0 );
		//new z-value
		pVal = stristr( pChr, "z=" );
		if( pVal != NULL )
			Front.Z = DoubleMinMax( atof( pVal + 2 ), -1.0, 1.0 );
		//repair position and front direction
		RepairPosFront( );
		//display front-direction
		sprintf( Text, "  front-direction is (%f, %f, %f)", Front.X, Front.Y, Front.Z );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//show cam-points
	if( strnicmp( Command, "camptshow", 9 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 9;
		for( ; *pChr == ' '; pChr++ );
		//display cam-points
		sprintf( Text, "  camera-points: %d", CamptsCnt );
		ConsoleInsertLine( Text );
		for( J = 0; J < CamptsCnt; J++ )
		{
			sprintf( Text, "  %3d: pos (%f, %f, %f) front (%f, %f, %f)", J,
				pCampts[J].Pos.X, pCampts[J].Pos.Y, pCampts[J].Pos.Z,
				pCampts[J].Front.X, pCampts[J].Front.Y, pCampts[J].Front.Z );
			ConsoleInsertLine( Text );
		}
		//short command
		return false;
	}

	//clear cam-points
	if( strnicmp( Command, "camptclr", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//clear cam-points
		CamptsCnt = 0;
		pCampts = (tCampt *)realloc( pCampts, 0 );
		//display message
		ConsoleInsertLine( "  camera-points cleared" );
		//short command
		return false;
	}

	//set cam-point n
	if( strnicmp( Command, "camptset", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//get cam-point number
		J = atoi( pChr );
		//jump over number and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//error
		if( J < 0 || J > CamptsCnt - 1 )
		{
			sprintf( Text, "  camera-point %d does not exist", J );
			ConsoleInsertLine( Text );
		}
		//set cam-point
		else
		{
			//new x-value for pos
			pVal = stristr( pChr, "px=" );
			if( pVal != NULL )
				pCampts[J].Pos.X = DoubleMinMax( atof( pVal + 3 ), PosMinX, PosMaxX );
			//new y-value for pos
			pVal = stristr( pChr, "py=" );
			if( pVal != NULL )
				pCampts[J].Pos.Y = DoubleMinMax( atof( pVal + 3 ), PosMinY, PosMaxY );
			//new z-value for pos
			pVal = stristr( pChr, "pz=" );
			if( pVal != NULL )
				pCampts[J].Pos.Z = DoubleMinMax( atof( pVal + 3 ), PosMinZ, PosMaxZ );
			//new x-value for front
			pVal = stristr( pChr, "fx=" );
			if( pVal != NULL )
				pCampts[J].Front.X = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
			//new y-value for front
			pVal = stristr( pChr, "fy=" );
			if( pVal != NULL )
				pCampts[J].Front.Y = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
			//new z-value for front
			pVal = stristr( pChr, "fz=" );
			if( pVal != NULL )
				pCampts[J].Front.Z = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
			//repair position and front direction
			RepairPosFront( );
			//display message
			sprintf( Text, "  camera-point %d set to pos (%f, %f, %f) front (%f, %f, %f)", J,
				pCampts[J].Pos.X, pCampts[J].Pos.Y, pCampts[J].Pos.Z,
				pCampts[J].Front.X, pCampts[J].Front.Y, pCampts[J].Front.Z );
			ConsoleInsertLine( Text );
		}
		//short command
		return false;
	}

	//insert cam-point n
	if( strnicmp( Command, "camptins", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//get waypoint number
		J = atoi( pChr );
		if( J < 0 )
			J = 0;
		if( J > CamptsCnt )
			J = CamptsCnt;
		//jump over number and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//insert cam-point
		CamptsCnt++;
		pCampts = (tCampt *)realloc( pCampts, CamptsCnt * sizeof( tCampt ) );
		for( K = CamptsCnt - 1; K > J; K-- )
		{
			pCampts[K].Pos = pCampts[K-1].Pos;
			pCampts[K].Front = pCampts[K-1].Front;
		}
		pCampts[J].Pos = Pos;
		pCampts[J].Front = Front;
		//new x-value for pos
		pVal = stristr( pChr, "px=" );
		if( pVal != NULL )
			pCampts[J].Pos.X = DoubleMinMax( atof( pVal + 3 ), PosMinX, PosMaxX );
		//new y-value for pos
		pVal = stristr( pChr, "py=" );
		if( pVal != NULL )
			pCampts[J].Pos.Y = DoubleMinMax( atof( pVal + 3 ), PosMinY, PosMaxY );
		//new z-value for pos
		pVal = stristr( pChr, "pz=" );
		if( pVal != NULL )
			pCampts[J].Pos.Z = DoubleMinMax( atof( pVal + 3 ), PosMinZ, PosMaxZ );
		//new x-value for front
		pVal = stristr( pChr, "fx=" );
		if( pVal != NULL )
			pCampts[J].Front.X = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
		//new y-value for front
		pVal = stristr( pChr, "fy=" );
		if( pVal != NULL )
			pCampts[J].Front.Y = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
		//new z-value for front
		pVal = stristr( pChr, "fz=" );
		if( pVal != NULL )
			pCampts[J].Front.Z = DoubleMinMax( atof( pVal + 3 ), -1.0, 1.0 );
		//repair position and front direction
		RepairPosFront( );
		//display message
		sprintf( Text, "  camera-point %d inserted and set to pos (%f, %f, %f) front (%f, %f, %f)", J,
			pCampts[J].Pos.X, pCampts[J].Pos.Y, pCampts[J].Pos.Z,
			pCampts[J].Front.X, pCampts[J].Front.Y, pCampts[J].Front.Z );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//delete cam-point n
	if( strnicmp( Command, "camptdel", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//get cam-point number
		J = atoi( pChr );
		//jump over number and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//error
		if( J < 0 || J > CamptsCnt - 1 )
		{
			sprintf( Text, "  camera-point %d does not exist", J );
			ConsoleInsertLine( Text );
		}
		//delete cam-point
		else
		{
			CamptsCnt--;
			for( K = J; K < CamptsCnt; K++ )
			{
				pCampts[K].Pos = pCampts[K+1].Pos;
				pCampts[K].Front = pCampts[K+1].Front;
			}
			pCampts = (tCampt *)realloc( pCampts, CamptsCnt * sizeof( tCampt ) );
			//display message
			sprintf( Text, "  camera-point %d deleted", J );
			ConsoleInsertLine( Text );
		}
		//short command
		return false;
	}

	//cam-point-speed
	if( strnicmp( Command, "camptspeed", 10 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 10;
		for( ; *pChr == ' '; pChr++ );
		//new value
		if( *pChr != 0 )
			CamptSpeed = DoubleMinMax( atof( pChr ), -CamptSpeedRange, CamptSpeedRange );
		//display speed
		sprintf( Text, "  camera-point-move-speed is now %f waypoints per second", CamptSpeed );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//cam-point-reference-time
	if( strnicmp( Command, "camptreftime", 12 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 12;
		for( ; *pChr == ' '; pChr++ );
		//new value
		if( *pChr != 0 )
		{
			//current time
			if( strnicmp( pChr, "time", 4 ) == 0 )
				CamptRefTime = TimeGet( );
			//value in seconds
			else
				CamptRefTime = DoubleMinMax( atof( pChr ), -TimeRange, TimeRange );
		}
		//display message
		sprintf( Text, "  reference time for camera-point-movement is %f", CamptRefTime );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//camera moving or not
	if( strnicmp( Command, "cammove", 7 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 7;
		for( ; *pChr == ' '; pChr++ );
		//toggle
		CamMoving = !CamMoving;
		//set on or off
		if( strnicmp( pChr, "on", 2 ) == 0 )
			CamMoving = true;
		if( strnicmp( pChr, "off", 3 ) == 0 )
			CamMoving = false;
		//display state
		if( CamMoving )
			ConsoleInsertLine( "  camera is now moving" );
		else
			ConsoleInsertLine( "  camera is now stopped" );
		//short command
		return false;
	}

	//console
	if( strnicmp( Command, "console", 7 ) == 0 )
	{
		//toggle console
		Console = !Console;
		//jump over spaces
		pChr = Command + 7;
		for( ; *pChr == ' '; pChr++ );
		//set on or off
		if( strnicmp( pChr, "on", 2 ) == 0 )
			Console = true;
		if( strnicmp( pChr, "off", 3 ) == 0 )
			Console = false;
		//display state
		if( Console )
			ConsoleInsertLine( "  console is now turned on" );
		else
			ConsoleInsertLine( "  console is now turned off" );
		//short command
		return false;
	}

	//background
	if( strnicmp( Command, "backgr", 6 ) == 0 )
	{
		//toggle background
		Back = !Back;
		//jump over spaces
		pChr = Command + 6;
		for( ; *pChr == ' '; pChr++ );
		//set on or off
		if( strnicmp( pChr, "on", 2 ) == 0 )
			Back = true;
		if( strnicmp( pChr, "off", 3 ) == 0 )
			Back = false;
		//display state
		if( Back )
			ConsoleInsertLine( "  background is now turned on" );
		else
			ConsoleInsertLine( "  background is now turned off" );
		//short command
		return false;
	}

	//antialiasing
	if( strnicmp( Command, "antialias", 9 ) == 0 )
	{
		//toggle antialiasing
		AntiAlias = !AntiAlias;
		//jump over spaces
		pChr = Command + 9;
		for( ; *pChr == ' '; pChr++ );
		//set on or off
		if( strnicmp( pChr, "on", 2 ) == 0 )
			AntiAlias = true;
		if( strnicmp( pChr, "off", 3 ) == 0 )
			AntiAlias = false;
		//display state
		if( AntiAlias )
			ConsoleInsertLine( "  antialiasing is now turned on" );
		else
			ConsoleInsertLine( "  antialiasing is now turned off" );
		//short command
		return false;
	}

	//load background from bitmap
	if( strnicmp( Command, "backloadbmp", 11 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 11;
		for( ; *pChr == ' '; pChr++ );
		//load new background
		FuRes = LoadTexBmpPieces( pChr, BackTexFirstNo, BackTexCntX, BackTexCntY, BackTexSizeX, BackTexSizeY );
		//display message
		if( FuRes )
			ConsoleInsertLine( "  background loaded" );
		else
			ConsoleInsertLine( "  background not loaded" );
		//long command
		return true;
	}

	//load background from jpeg
	if( strnicmp( Command, "backloadjpg", 11 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 11;
		for( ; *pChr == ' '; pChr++ );
		//load new background
		FuRes = LoadTexJpgPieces( pChr, BackTexFirstNo, BackTexCntX, BackTexCntY, BackTexSizeX, BackTexSizeY );
		//display message
		if( FuRes )
			ConsoleInsertLine( "  background loaded" );
		else
			ConsoleInsertLine( "  background not loaded" );
		//long command
		return true;
	}

	//unload background
	if( strnicmp( Command, "backunload", 10 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 10;
		for( ; *pChr == ' '; pChr++ );
		//unload background-textures
		for( I = BackTexFirstNo; I < BackTexFirstNo + BackTexCnt; I++ )
			UnloadTexture( I );
		//display message
		ConsoleInsertLine( "  background unloaded" );
		//long command
		return true;
	}

	//load pixel texture from bitmap
	if( strnicmp( Command, "pixtexloadbmp", 13 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 13;
		for( ; *pChr == ' '; pChr++ );
		//load new texture
		FuRes = LoadTexBmp( pChr, PixTexNo, PixTexSizeX, PixTexSizeY );
		//display message
		if( FuRes )
			sprintf( Text, "  pixel texture loaded" );
		else
			sprintf( Text, "  pixel texture not loaded" );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}
	
	//load pixel texture from jpeg
	if( strnicmp( Command, "pixtexloadjpg", 13 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 13;
		for( ; *pChr == ' '; pChr++ );
		//load new texture
		FuRes = LoadTexJpg( pChr, PixTexNo, PixTexSizeX, PixTexSizeY );
		//display message
		if( FuRes )
			sprintf( Text, "  pixel texture loaded" );
		else
			sprintf( Text, "  pixel texture not loaded" );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//unload texture
	if( strnicmp( Command, "pixtexunload", 12 ) == 0 )
	{
		//unload texture
		UnloadTexture( PixTexNo );
		//display message
		sprintf( Text, "  pixel texture unloaded" );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//set time
	if( strnicmp( Command, "time", 4 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 4;
		for( ; *pChr == ' '; pChr++ );
		//new value
		if( *pChr != 0 )
			TimeSet( DoubleMinMax( atof( pChr ), -TimeRange, TimeRange ) );
		//display message
		sprintf( Text, "  current time is %f", TimeGet( ) );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//load a script-file
	if( strnicmp( Command, "scriptload", 10 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 10;
		for( ; *pChr == ' '; pChr++ );
		//load script
		I = LoadScript( pChr );
		//display number of lines loaded
		if( I < 0 )
			sprintf( Text, "  script could not be loaded" );
		else
			sprintf( Text, "  script loaded (%d lines)", I );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//unload script
	if( strnicmp( Command, "scriptunload", 12 ) == 0 )
	{
		//empty script-buffer
		ScriptBuffer = (char*)realloc( ScriptBuffer, 1 );
		ScriptBuffer[0] = 0;
		//display message
		ConsoleInsertLine( "  script unloaded" );
		//long command
		return true;
	}

	//go on with script execution
	if( strnicmp( Command, "scriptstart", 11 ) == 0 )
	{
		//set script-mode to running
		ScriptMode = 'R';
		//display message
		ConsoleInsertLine( "  script execution goes on" );
		//long command
		return true;
	}

	//halt script execution
	if( strnicmp( Command, "scriptstop", 10 ) == 0 )
	{
		//set script-mode to stopped
		ScriptMode = 'S';
		//display message
		ConsoleInsertLine( "  script execution halted" );
		//long command
		return true;
	}

	//pause script absolute
	if( strnicmp( Command, "scriptwaitabs", 13 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 13;
		for( ; *pChr == ' '; pChr++ );
		//get value
		if( *pChr != 0 )
			ScriptWaitTime = DoubleMinMax( atof( pChr ), -TimeRange, TimeRange );
		//set script-mode to waiting
		ScriptMode = 'W';
		//display message
		sprintf( Text, "  current time is %f", TimeGet( ) );
		ConsoleInsertLine( Text );
		sprintf( Text, "  script execution halted till time %f", ScriptWaitTime );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//pause script relative
	if( strnicmp( Command, "scriptwaitrel", 13 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 13;
		for( ; *pChr == ' '; pChr++ );
		//get value
		if( *pChr != 0 )
			ScriptWaitTime = DoubleMinMax( atof( pChr ) + TimeGet( ), 0.0, TimeRange );
		//set script-mode to waiting
		ScriptMode = 'W';
		//display message
		sprintf( Text, "  current time is %f", TimeGet( ) );
		ConsoleInsertLine( Text );
		sprintf( Text, "  script execution halted till time %f", ScriptWaitTime );
		ConsoleInsertLine( Text );
		//long command
		return true;
	}

	//set entire room to a color
	if( strnicmp( Command, "colorset", 8 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 8;
		for( ; *pChr == ' '; pChr++ );
		//get color
		Color = (unsigned char)LongMinMax( atoi( pChr ), 0, 255 );
		//paint walls
		FrameSet( Color );
		//display message
		sprintf( Text, "  walls painted in color %d", Color );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//set entire room to random colors
	if( strnicmp( Command, "colorrandom", 11 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 11;
		for( ; *pChr == ' '; pChr++ );
		//get minimum color
		if( *pChr == 0 )
			Color = 0;
		else
			Color = (unsigned char)LongMinMax( atoi( pChr ), 0, 255 );
		//jump over minimum color and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//get maximum color
		if( *pChr == 0 )
			Color2 = 255;
		else
			Color2 = (unsigned char)LongMinMax( atoi( pChr ), 0, 255 );
		//paint walls in random colors
		FrameRandomize( Color, Color2 );
		//display message
		sprintf( Text, "  walls painted in random colors from %d to %d", Color, Color2 );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//write text to the room
	if( strnicmp( Command, "textwrite", 9 ) == 0 )
	{
		//jump over spaces
		pChr = Command + 9;
		for( ; *pChr == ' '; pChr++ );
		//get pixel
		Pixel = Txt2Pixel3D( pChr );
		//jump over pixel and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//get direction
		Dir = Txt2Dir( pChr );
		//jump over direction and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//get foreground color
		ForeColor = (unsigned short)LongMinMax( atoi( pChr ), 0, 256 ); //256 is transparent
		//jump over foreground color and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//get background color
		BackColor = (unsigned short)LongMinMax( atoi( pChr ), 0, 256 ); //256 is transparent
		//jump over background color and over spaces
		for( ; *pChr != ' ' && *pChr != 0; pChr++ );
		for( ; *pChr == ' '; pChr++ );
		//write text
		FrameTextOut( pChr, Pixel, Dir, ForeColor, BackColor );
		//display message
		sprintf( Text, "  wrote %d characters to ", strlen( pChr ) );
		Pixel3D2Txt( Pixel, Text + strlen( Text ) );
		strcat( Text, " in direction " );
		Dir2Txt( Dir, Text + strlen( Text ) );
		sprintf( Text + strlen( Text ), " in color %d on color %d", ForeColor, BackColor );
		ConsoleInsertLine( Text );
		//short command
		return false;
	}

	//unknown command
	ConsoleInsertLine( "  unknown command (type \"help\" for list of commands)" );
	//long command
	return true;
}

//process a command from the script-buffer
//return true for "long" commands, false for "short" commands
static int ProcessScriptBufferCommand( )
{
	//local variables
	char * pChr, * pNl, * pCommand;
	char ConsoleLine[ConsoleTextSizeX];
	int LongCommand;

	//buffer empty?
	if( ScriptBuffer[0] == 0 )
		//long command
		return true;

	//jump over spaces and newlines
	for( pChr = ScriptBuffer; *pChr == ' ' || *pChr == '\n'; pChr++ );

	//end of buffer?
	if( *pChr == 0 )
	{
		//empty buffer
		ScriptBuffer[0] = 0;
		ScriptBuffer = (char *)realloc( ScriptBuffer, 1 );
		//long command
		return true;
	}

	//search for next newline
	pNl = strchr( pChr, '\n' );
	if( pNl == NULL ) //not found
	{
		//put command in extra-buffer
		pCommand = strdup( pChr );
		//empty script-buffer
		ScriptBuffer = (char *)realloc( ScriptBuffer, 1 );
		ScriptBuffer[0] = 0;
	}
	else //found
	{
		//put command in extra-buffer
		*pNl = 0; //terminate string before newline
		pCommand = strdup( pChr );
		//remove command from buffer and resize buffer
		pChr = pNl + 1;
		memmove( ScriptBuffer, pChr, strlen( pChr ) + 1 );
		ScriptBuffer = (char *)realloc( ScriptBuffer, strlen( ScriptBuffer ) + 1 );
	}

	//display command
	strcpy( ConsoleLine, "[script] " );
	strncat( ConsoleLine, pCommand, sizeof( ConsoleLine ) - strlen( ConsoleLine ) - 1 );
	ConsoleInsertLine( ConsoleLine );
	//process command (ATTENTION: could append something to ScriptBuffer)
	LongCommand = ProcessCommand( pCommand );
	//free command-string
	free( pCommand );

	//long command?
	return LongCommand;
}

//set contents of script buffer
void ScriptBufferSet( char * pTxt )
{
	if( ScriptBuffer != NULL )
		free( ScriptBuffer );
	ScriptBuffer = NULL;
	if( pTxt != NULL )
		ScriptBuffer = strdup( pTxt );
}

//cleanup script buffer
void ScriptBufferCleanup( )
{
	if( ScriptBuffer != NULL )
		free( ScriptBuffer );
	ScriptBuffer = NULL;
}

//process a few command from the script-buffer
void ProcessScriptBufferCommands( )
{
	//local variables
	long I;

	//script-mode is waiting
	if( ScriptMode == 'W' )
	{
		//time reached => switch to running
		if( TimeGet( ) >= ScriptWaitTime )
			ScriptMode = 'R';
	}

	//script-mode is running
	if( ScriptMode == 'R' )
	{
		//process max MaxScriptCommandsAtSameTime commands
		for( I = 0; I < MaxScriptCommandsAtSameTime; I++ )
			//process command
			if( ProcessScriptBufferCommand( ) )
				//long command => no more commands
				break;
	}
}
