/* TroiaPreview
 * version 0.2 date 2004-09-01
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <math.h>

#include "Tools.h"
#include "Vector.h"

//set vector
void SetVector( tVector * pDest, double X, double Y, double Z )
{
	pDest->X = X;
	pDest->Y = Y;
	pDest->Z = Z;
}

//vector addition
void Add( tVector * pDest, tVector Src1, tVector Src2 )
{
	pDest->X = Src1.X + Src2.X;
	pDest->Y = Src1.Y + Src2.Y;
	pDest->Z = Src1.Z + Src2.Z;
}

//vector subtraction
void Sub( tVector * pDest, tVector Src1, tVector Src2 )
{
	pDest->X = Src1.X - Src2.X;
	pDest->Y = Src1.Y - Src2.Y;
	pDest->Z = Src1.Z - Src2.Z;
}

//scale a vector
void Scale( tVector * pV, double Factor )
{
	pV->X *= Factor;
	pV->Y *= Factor;
	pV->Z *= Factor;
}

//linear-combine two vectors
void LinearCombine( tVector * pDest, double Factor1, tVector Src1, double Factor2, tVector Src2 )
{
	pDest->X = Factor1 * Src1.X + Factor2 * Src2.X;
	pDest->Y = Factor1 * Src1.Y + Factor2 * Src2.Y;
	pDest->Z = Factor1 * Src1.Z + Factor2 * Src2.Z;
}

//length of a vector
double Len( tVector V )
{
	return sqrt( V.X * V.X + V.Y * V.Y + V.Z * V.Z );
}

//normalize a vector
void Normalize( tVector * pV )
{
	double L;
	L = Len( *pV );
	if( L == 0.0 )
		pV->X = 0.0, pV->Y = 0.0, pV->Z = 1.0;
	else
		Scale( pV, 1.0 / L );
}

//scalar product
double ScalarProduct( tVector Vec1, tVector Vec2 )
{
	return Vec1.X * Vec2.X + Vec1.Y * Vec2.Y + Vec1.Z * Vec2.Z;
}

//angle between two vectors
double VectorAngle( tVector Vec1, tVector Vec2 )
{
	return acos( ScalarProduct( Vec1, Vec2 ) / ScalarProduct( Vec2, Vec2 ) );
}

//vector product
void VectorProduct( tVector * pDest, tVector Vec1, tVector Vec2 )
{
	pDest->X = Vec1.Y * Vec2.Z - Vec1.Z * Vec2.Y;
	pDest->Y = Vec1.Z * Vec2.X - Vec1.X * Vec2.Z;
	pDest->Z = Vec1.X * Vec2.Y - Vec1.Y * Vec2.X;
}

//rotate vector V around vector A
void Rotate( tVector * pV, tVector A, double Angle )
{
	//local variables
	tVector V;
	double R;

	//normalize axis A
	Normalize( &A );

	//turn around X-Axis
	V = *pV;
	R = Angle * A.X;
	pV->Y = cos( R ) * V.Y + sin( R ) * V.Z;
	pV->Z = -sin( R ) * V.Y + cos( R ) * V.Z;

	//turn around Y-Axis
	V = *pV;
	R = Angle * A.Y;
	pV->X = cos( R ) * V.X - sin( R ) * V.Z;
	pV->Z = sin( R ) * V.X + cos( R ) * V.Z;

	//turn around Z-Axis
	V = *pV;
	R = Angle * A.Z;
	pV->X = cos( R ) * V.X + sin( R ) * V.Y;
	pV->Y = -sin( R ) * V.X + cos( R ) * V.Y;
}
